"""
Eraser class
"""
import asyncio
import platform
import sys
from getpass import getpass
from time import sleep
from typing import Any, List, Set, Union

from telethon import TelegramClient, hints
from telethon.errors import SessionPasswordNeededError
from telethon.network import ConnectionTcpAbridged
from telethon.tl.functions.channels import DeleteMessagesRequest
from telethon.tl.functions.messages import (
    DeleteMessagesRequest as DeleteMessagesRequestFromUser,
)
from telethon.tl.functions.messages import SearchRequest
from telethon.tl.types import Channel, InputMessagesFilterEmpty, InputUserSelf, User
from telethon.utils import get_display_name

from .__version__ import VERSION
from .exceptions import TgEraserException
from .utils import cast_to_int, chunks, print_header, sprint

_ = Any, List, Set, Union

loop = asyncio.get_event_loop()


async def async_input(prompt: str) -> str:
    """
    Python's ``input()`` is blocking, which means the event loop we set
    above can't be running while we're blocking there. This method will
    let the loop run while we wait for input.
    """
    print(prompt, end="", flush=True)
    return (await loop.run_in_executor(None, sys.stdin.readline)).rstrip()


class Eraser(TelegramClient):  # type: ignore
    """
    Subclass of TelegramClient class
    """

    def __init__(self: TelegramClient, **kwargs: Any) -> None:
        super().__init__(
            session=kwargs["session_name"],
            api_id=kwargs["api_id"],
            api_hash=kwargs["api_hash"],
            connection=ConnectionTcpAbridged,
            device_model=platform.uname().system,
            system_version=platform.uname().release,
            app_version=VERSION,
        )

        self.__limit = kwargs["limit"]  # Limit to retrieve the top dialogs
        self.__peers = kwargs["peers"].split(",") if kwargs["peers"] else []
        self.__dialogs = kwargs["dialogs"]
        self.__channels = kwargs["channels"]
        self.__entities: List[hints.Entity] = []
        self.__display_name = ""
        self.__messages_to_delete: Set[int] = set()

        # Check connection to the server
        print("Connecting to Telegram servers...")
        try:
            loop.run_until_complete(self.connect())
        except ConnectionError:
            print("Initial connection failed. Retrying...")
            loop.run_until_complete(self.connect())

        # Check authorization
        if not loop.run_until_complete(self.is_user_authorized()):
            print("First run. Sending code request...")
            if kwargs.get("user_phone", None):
                user_phone = kwargs["user_phone"]
            else:
                user_phone = input("Enter your phone: ")
            loop.run_until_complete(self.sign_in(user_phone))

            self_user = None
            while self_user is None:
                code = input("Enter the code you just received: ")
                try:
                    self_user = loop.run_until_complete(self.sign_in(code=code))

                # Two-step verification may be enabled
                except SessionPasswordNeededError:
                    password = getpass(
                        "Two step verification is enabled. "
                        "Please enter your password: "
                    )

                    self_user = loop.run_until_complete(self.sign_in(password=password))

    async def run(self) -> None:
        """
        Runs deletion of messages from peer
        """
        if self.__peers:
            for peer in self.__peers:
                if peer.isdigit():
                    peer = cast_to_int(peer, f"peer: {peer}")
                try:
                    self.__entities.append(await self.get_entity(peer))
                except ValueError:
                    raise TgEraserException(
                        f"Specified entity '{peer}' can't be found."
                    )
        else:
            await self.__get_entity()
        for entity in self.__entities:
            self.__display_name = get_display_name(entity)
            self.__messages_to_delete.update(
                msg.id for msg in await self.__get_messages(entity)
            )
            await self.__delete_messages_from_peer(entity)
            self.__messages_to_delete.clear()

    async def __get_entity(self) -> None:
        """
        Returns chosen peer
        """
        entities = [d.entity for d in await self.get_dialogs(limit=self.__limit)]

        entities = [
            entity
            for entity in entities
            if isinstance(entity, User if self.__dialogs else Channel)
        ]
        if not self.__dialogs and not self.__channels:
            entities = [entity for entity in entities if entity.megagroup]

        if not entities:
            raise TgEraserException("You aren't joined to any chat.")

        print_header("Dialogs")
        for i, entity in enumerate(entities, start=1):
            sprint(f"{i}. {get_display_name(entity)}\t | {entity.id}")

        num = cast_to_int(await async_input("\nChoose peer: "), "peer")
        self.__entities = [entities[int(num) - 1]]
        self.__display_name = get_display_name(self.__entities[0])
        print(self.__display_name)
        print("Chosen: " + self.__display_name)

    async def __delete_messages_from_peer(self, entity: hints.Entity) -> None:
        """
        Method deletes messages
        """
        messages_to_delete = list(self.__messages_to_delete)
        print_header(
            f"Delete {len(messages_to_delete)} of my messages in chat {self.__display_name}"
        )
        for chunk_data in chunks(
            messages_to_delete, 100
        ):  # Because we can delete only 100 messages per request (Telegram API restrictions)
            if self.__dialogs:
                result = await self(
                    DeleteMessagesRequestFromUser(chunk_data, revoke=True)
                )
            else:
                result = await self(DeleteMessagesRequest(entity, chunk_data))
            if result.pts_count:
                print(f"Number of deleted messages: {result.pts_count}")
            sleep(1)
        print_header("Erasing is finished.")

    async def __get_messages(
        self,
        entity: hints.Entity,
        limit: int = 100,
        offset_id: int = 0,
        max_id: int = 0,
        min_id: int = 0,
    ) -> List[Any]:
        print_header(f"Getting messages from {self.__display_name}...")
        add_offset = 0
        messages: List[Any] = []

        while True:
            result = await self(
                SearchRequest(
                    peer=entity,
                    q="",
                    filter=InputMessagesFilterEmpty(),
                    min_date=None,
                    max_date=None,
                    offset_id=offset_id,
                    add_offset=add_offset,
                    limit=limit,
                    max_id=max_id,
                    min_id=min_id,
                    from_id=InputUserSelf(),
                    hash=0,
                )
            )

            if result.messages:
                print(
                    f"Received: {len(result.messages)} messages. Offset: {add_offset}."
                )
                messages.extend(result.messages)
                add_offset += len(result.messages)
                sleep(1)
            else:
                print("It's finished because it met end of chat.")
                return messages



"""
small Python functions and classes which make common patterns shorter and easier
"""
import json
import os
import re
import sys
from typing import Any, Dict, Iterable, List, Optional, Set, Union

import yaml

from .exceptions import TgEraserException

_ = Any, Dict, Iterable, List, Optional, Set, Union


def chunks(input_list: List[Any], num: int) -> Iterable[Any]:
    """Splits list input_list into chunks of size n. Returns generator"""
    for i in range(0, len(input_list), num):
        yield input_list[i : i + num]


def sprint(string: str, *args: Any, **kwargs: Any) -> None:
    """Safe Print (handle UnicodeEncodeErrors on some terminals)"""
    try:
        print(string, *args, **kwargs)
    except UnicodeEncodeError:
        string = string.encode("utf-8", errors="ignore").decode(
            "ascii", errors="ignore"
        )
        print(string, *args, **kwargs)


def print_header(title: str) -> None:
    """Helper function to print titles to the console more nicely"""
    sprint("\n")
    sprint("=={}==".format("=" * len(title)))
    sprint("= {} =".format(title))
    sprint("=={}==".format("=" * len(title)))


def get_env(name: str, message: str, cast: Any = str) -> Any:
    """Helper to get environment variables interactively"""
    if name in os.environ:
        return os.environ[name]
    while True:
        value = input(message)
        try:
            return cast(value)
        except ValueError as err:
            print(err, file=sys.stderr)


def cast_to_int(num: str, name: str) -> int:
    """check if a string represents an int"""
    try:
        return int(num)
    except ValueError as err:
        raise TgEraserException(f"Error: '{name}' should be integer.") from err


def get_credentials(args: Dict[str, Any]) -> Dict[str, str]:
    """Presents credentials dict from specified source"""

    path_to_file = os.path.abspath(os.path.expanduser(args["--input-file"]))
    path_to_directory = os.path.dirname(path_to_file) + "/"
    os.makedirs(path_to_directory, exist_ok=True)

    creds = {}  # type: Dict[str, Any]
    if args["--json"]:
        creds = get_credentials_from_json(
            args["--json"], path_to_directory, args["<session_name>"]
        )
    elif args["--environment-variables"]:
        creds = get_credentials_from_env(path_to_directory)
    else:
        if not os.path.exists(path_to_file):
            answer = None
            while answer not in ("yes", "no", "y", "n"):
                answer = input("Do you want create file? (y/n): ").lower()
                if answer in ("yes", "y"):
                    creds = create_credential_file(path_to_file, path_to_directory)
                elif answer in ("no", "n"):
                    exit(1)
                else:
                    print("Please enter yes or no.")
        else:
            creds = get_credentials_from_yaml(
                path_to_file, path_to_directory, args["<session_name>"]
            )

    return creds


def get_credentials_from_yaml(
    path_to_file: str, path_to_directory: str, session_name: Optional[str] = None
) -> Dict[str, str]:
    """
    Returns credentials and certain session from YAML file
    """
    creds = yaml.load(open(path_to_file, "r"), Loader=yaml.SafeLoader)
    check_credentials_dict(creds)

    if session_name:
        for i, cred in enumerate(creds["sessions"]):
            if cred["session_name"] == session_name:
                creds["sessions"][i]["session_name"] = (
                    path_to_directory + session_name + ".session"
                )
                print("Session file: " + creds["sessions"][i]["session_name"])
                return {**creds["api_credentials"], **creds["sessions"][i]}

        raise TgEraserException(
            f"It can't find '{session_name}' session in credentials file."
        )

    else:
        print_header("Sessions")
        for i, cred in enumerate(creds["sessions"], start=1):
            sprint(
                "{0}. {1}\t | {2}".format(
                    i,
                    cred["session_name"],
                    cred.get("user_phone", "Phone number wasn't specified"),
                )
            )

        num = int(input("\nChoose session: ")) - 1
        print("Chosen: " + creds["sessions"][num]["session_name"] + "\n")

        creds["sessions"][num]["session_name"] = (
            path_to_directory + creds["sessions"][num]["session_name"] + ".session"
        )
        return {**creds["api_credentials"], **creds["sessions"][num]}


def create_credential_file(path: str, directory: str) -> Dict[str, str]:
    """
    creates credential YAML file
    """
    phone_pattern = re.compile(
        r"(([+][(]?[0-9]{1,3}[)]?)|([(]?[0-9]{4}[)]?))"
        r"\s*[)]?[-\s\.]?[(]?[0-9]{1,3}[)]?([-\s\.]?[0-9]{3})"
        r"([-\s\.]?[0-9]{3,4})"
    )
    credentials = {
        "api_credentials": {"api_id": "", "api_hash": ""},
        "sessions": [{"session_name": "", "user_phone": ""}],
    }  # type: Dict[str, Any]

    credentials["api_credentials"]["api_id"] = cast_to_int(
        input("Enter api_id: "), "api_id"
    )
    credentials["api_credentials"]["api_hash"] = input("Enter api_hash: ")

    credentials["sessions"][0]["session_name"] = input("Enter session_name: ")

    i: int = 0
    while True:
        credentials["sessions"][0]["user_phone"] = input(
            "Enter user_phone (+1234567890): "
        )
        if not phone_pattern.search(credentials["sessions"][0]["user_phone"]):
            print("Incorrect phone number. Try again.")
        else:
            break

        i += 1
        if i == 3:
            raise TgEraserException("Incorrect phone number. Exiting...")

    with open(path, "w") as yaml_file:
        yaml.dump(credentials, yaml_file, default_flow_style=False)

    print(f"Credentials file is created ({path}).")

    credentials["sessions"][0]["session_name"] = (
        directory + credentials["sessions"][0]["session_name"] + ".session"
    )

    return {**credentials["api_credentials"], **credentials["sessions"][0]}


def get_credentials_from_json(
    json_str: str, path: str, session_name: str
) -> Dict[str, Any]:
    """
    Returns credentials and certain session from JSON string
    """
    creds = json.loads(json_str)
    check_credentials_dict(creds)

    if session_name:
        for i, cred in enumerate(creds["sessions"]):
            if cred["session_name"] == session_name:
                creds["sessions"][i]["session_name"] = path + session_name + ".session"
                print("Session file: " + creds["sessions"][i]["session_name"])
                return {**creds["api_credentials"], **creds["sessions"][i]}

        raise TgEraserException(
            f"It can't find '{session_name}' session in credentials file."
        )
    else:
        print_header("Sessions")
        for i, cred in enumerate(creds["sessions"], start=1):
            sprint(f"{i}. {cred['session_name']}\t | {cred['user_phone']}")

        num = int(input("\nChoose session: ")) - 1
        print("Chosen: " + creds["sessions"][num]["session_name"] + "\n")

        creds["sessions"][num]["session_name"] = (
            path + creds["sessions"][num]["session_name"] + ".session"
        )
        return {**creds["api_credentials"], **creds["sessions"][num]}


def get_credentials_from_env(path: str) -> Dict[str, Any]:
    """Gets credentials from environment variables"""
    api_id = get_env("TG_API_ID", "Enter your API ID: ", int)
    api_hash = get_env("TG_API_HASH", "Enter your API hash: ")
    session = get_env("TG_SESSION", "Enter session name: ")
    session = path + session + ".session"
    return {"api_id": api_id, "api_hash": api_hash, "session_name": session}


def check_credentials_dict(creds: Dict[str, Any]) -> None:
    """Checks basic structure of credentials dictionary"""
    if not creds["api_credentials"]:
        raise TgEraserException("Credentials file doesn't contain 'api_credentials'.")
    if not creds["api_credentials"]["api_id"]:
        raise TgEraserException("Credentials file doesn't contain 'api_id'.")
    if not creds["api_credentials"]["api_hash"]:
        raise TgEraserException("Credentials file doesn't contain 'api_hash'.")
    if not creds["sessions"]:
        raise TgEraserException("Credentials file doesn't contain sessions.")
    if not creds["sessions"][0]["session_name"]:
        raise TgEraserException("Credentials file doesn't contain session_name.")



"""
Tool deletes all your messages from chat/channel/dialog on Telegram.

Usage:
    tgeraser [ (session <session_name>) -cdl NUM [ -i FILEPATH | -j DICT | -e ] -p PEER_ID ] | [ -k ]
    tgeraser session <session_name> -p ID -t STRING
    tgeraser -h | --help
    tgeraser --version

Options:
    -i --input-file FILEPATH    Specify YAML file that contains credentials. [default: ~/.tgeraser/credentials.yml]
    -j --json DICT              Specify json string that contains credentials (double quotes must be escaped).
    -e --environment-variables  Get credentials from environment variables (TG_API_ID, TG_API_HASH, TG_SESSION).
    -d --dialogs                List only Dialogs (Chats by default).
    -c --channels               List only Channels (Chats by default).
    -p --peers PEER_ID          Specify certain peers by comma (chat/channel/dialog).
    -l --limit NUM              Show specified number of recent chats.
    -t --time-period STRING     Specify period for infinite loop to run messages deletion every X seconds/minutes/hours/days/weeks.
                                Example: --time-period "3*days" OR --time-period "5*seconds"
    -k --kill                   Kill existing background tgeraser processes (only for Unix-like OS).
    -h --help                   Show this screen.
    --version                   Show version.

"""  # pylint: disable=line-too-long

import asyncio
import os
import signal
import subprocess
import sys
import time
import traceback

from docopt import docopt

from .eraser import Eraser
from .__version__ import VERSION
from .exceptions import TgEraserException
from .utils import cast_to_int, get_credentials

loop = asyncio.get_event_loop()


def entry() -> None:
    """
    Entry function
    """
    arguments = docopt(__doc__, version=VERSION)
    if arguments["--limit"]:
        arguments["--limit"] = cast_to_int(arguments["--limit"], "limit")
    if arguments["--time-period"]:
        periods = {
            "seconds": 1,
            "minutes": 60,
            "hours": 3600,
            "days": 86400,
            "weeks": 604800,
        }
        period = arguments["--time-period"].split("*")
        if period[1] not in periods:
            raise TgEraserException("Time period is specified incorrectly.")

    if arguments["--kill"]:
        if os.name != "posix":
            raise TgEraserException("You can't use '--kill' option on Windows.")
        cmd = subprocess.Popen(["ps", "-A"], stdout=subprocess.PIPE)
        out = cmd.communicate()[0]
        current_pid = os.getpid()
        for line in out.splitlines():
            if b"tgeraser" in line:
                pid = int(line.split(None, 1)[0])
                if current_pid != pid:
                    os.kill(pid, signal.SIGKILL)
                    print("Process " + str(pid) + " successfully killed.")
        sys.exit(0)

    try:
        credentials = get_credentials(arguments)

        kwargs = {
            **credentials,
            "dialogs": arguments["--dialogs"],
            "channels": arguments["--channels"],
            "peers": arguments["--peers"],
            "limit": arguments["--limit"],
        }

        client = Eraser(**kwargs)
        while True:
            loop.run_until_complete(client.run())
            if arguments["--time-period"]:
                print(
                    "\n({0})\tNext erasing will be in {1} {2}.".format(
                        time.strftime("%Y-%m-%d, %H:%M:%S", time.gmtime()),
                        period[0],
                        period[1],
                    )
                )
                time.sleep(int(period[0]) * periods[period[1]])
            else:
                break
        client.disconnect()
        loop.close()
    except KeyboardInterrupt:
        print("\nExiting...")
    except Exception as err:  # pylint: disable=broad-except
        traceback.print_tb(err.__traceback__)


if __name__ == "__main__":
    entry()
